#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Contains one single class that simulates scenario-system (agent-environment) loops.

Contains one single class that simulates scenario-system (agent-environment) loops.
The system can be of three types:

- discrete-time deterministic
- continuous-time deterministic or stochastic
- discrete-time stochastic (to model Markov decision processes)

Remarks:

- All vectors are treated as of type [1, n]
- All buffers are treated as of type [L, n] where each row is a vector
- Buffers are updated from bottom to top

"""

import numpy as np

import regelum
from .utils import rg
from .system import System, ComposedSystem
from typing import Union, Optional
from abc import ABC

try:
    import casadi
except ModuleNotFoundError:
    from unittest.mock import MagicMock

    casadi = MagicMock()


class Simulator(regelum.RegelumBase, ABC):
    """Base class of Simulator."""

    def __init__(
        self,
        system: Union[System, ComposedSystem],
        state_init: Optional[np.ndarray] = None,
        action_init: Optional[np.ndarray] = None,
        time_final: Optional[float] = 1,
        max_step: Optional[float] = 1e-3,
        first_step: Optional[float] = 1e-6,
        atol: Optional[float] = 1e-5,
        rtol: Optional[float] = 1e-3,
    ):
        """Initialize an instance of Simulator.

        Args:
            system (Union[System, ComposedSystem]): A controlled system
                to be simulated
            state_init (Optional[np.ndarray], optional): Set initial
                state manually, defaults to None
            action_init (Optional[np.ndarray], optional): Set initial
                action manually, defaults to None
            time_final (Optional[float], optional): Time at which
                simulation ends, defaults to 1
            max_step (Optional[float], optional): Total duration of one
                simulation step, defaults to 1e-3
            first_step (Optional[float], optional): Used with
                integrators with changing simulation steps, defaults to
                1e-6
            atol (Optional[float], optional): Absolute tollerance of
                integrator, defaults to 1e-5
            rtol (Optional[float], optional): Relative tollerance of
                integrator, defaults to 1e-3
        """
        self.system = system
        assert hasattr(
            self.system, "system_type"
        ), "System must contain a system_type attribute"
        if self.system.system_type == "diff_eqn":
            assert (
                state_init is not None
            ), "Initial state for this simulator needs to be passed"

        self.time_final = time_final
        if state_init is None:
            self.state_init = self.initialize_init_state()
        else:
            self.state_init = state_init
        if action_init is None:
            self.action_init = self.initialize_init_action()
        else:
            self.action_init = action_init

        self.time = 0.0
        self.state = self.state_init
        self.observation_init = self.observation = self.get_observation(
            time=self.time, state=self.state_init, inputs=self.action_init
        )

        self.max_step = max_step
        self.atol = atol
        self.rtol = rtol
        self.first_step = first_step

        if self.system.system_type == "diff_eqn":
            self.ODE_solver = self.initialize_ode_solver()

    def receive_action(self, action):
        self.system.receive_action(action)

    def get_observation(self, time, state, inputs):
        return self.system.get_observation(time, state, inputs)

    def do_sim_step(self):
        """Do one simulation step and update current simulation data (time, system state and output)."""
        if self.system.system_type == "diff_eqn":
            try:
                self.ODE_solver.step()
            except RuntimeError:
                self.reset()
                return -1

            self.time = self.ODE_solver.t
            self.state = self.ODE_solver.y
            self.observation = self.get_observation(
                time=self.time, state=self.state, inputs=self.system.inputs
            )
        else:
            raise ValueError("Invalid system description")

    @apply_callbacks()
    def get_sim_step_data(self):
        return self.time, self.state, self.observation, self.get_simulation_metadata()

    def get_simulation_metadata(self): ...

    def reset(self):
        if self.system.system_type == "diff_eqn":
            self.ODE_solver = self.initialize_ode_solver()
            self.time = 0.0
            self.state = self.state_init
            self.observation = self.get_observation(
                time=self.time, state=self.state_init, inputs=self.action_init
            )
        else:
            self.time = 0.0
            self.observation = self.get_observation(
                time=self.time, state=self.state_init, inputs=self.system.inputs
            )

    def initialize_ode_solver(self):
        raise NotImplementedError("Not implemented ODE solver")

    def get_init_state_and_action(self):
        return self.state_init, rg.zeros((1, self.system.dim_inputs))

    def initialize_init_state(self):
        raise NotImplementedError(
            "Implement this method to initialize the initial state"
            + "if one is intended to be obtained during the runtime"
        )

    def initialize_init_action(self):
        return rg.zeros(self.system.dim_inputs)


class SciPy(Simulator):
    """Class for SciPy integrators."""

    def initialize_ode_solver(self):
        import scipy as sp

        ODE_solver = sp.integrate.RK45(
            self.system.compute_closed_loop_rhs,
            self.state,
            self.time_final,
            max_step=self.max_step,
            first_step=self.first_step,
            atol=self.atol,
            rtol=self.rtol,
        )
        return ODE_solver


class CasADi(Simulator):
    """Class for CasADi integrators."""

    class CasADiSolver:
        """Nested class to wrap casadi integrator into a uniform API."""

        def __init__(
            self,
            simulator: "CasADi",
        ):
            """Initialize a CasADiSolver object.

            Args:
                integrator (casadi.integrator): A CasADi integrator
                    object.
                time_final (float): The final time for the solver.
                step_size (float): The step size for the solver.
                action_init (np.array): The initial action for the
                    solver.
                system (System): The system object for the solver.
            """
            self.simulator = simulator
            self.integrator = simulator.integrator
            self.time_final = simulator.time_final
            self.step_size = simulator.max_step
            self.time = 0.0
            self.system = simulator.system

        def step(self):
            """Advance the solver by one step."""
            if self.time >= self.time_final:
                raise RuntimeError("An attempt to step with a finished solver")
            state_new = (
                self.integrator(x0=self.simulator.state, p=self.system.inputs)["xf"]
                .full()
                .T
            )

            self.time += self.step_size
            self.state = state_new

        @property
        def t(self):
            return self.time

        @property
        def y(self):
            return self.state

    def initialize_ode_solver(self):
        self.integrator = self.create_CasADi_integrator(self.max_step)
        assert self.time_final is not None and self.max_step is not None, (
            "Must specify time_final and max_step"
            + " in order to initialize CasADi solver"
        )
        ODE_solver = self.CasADiSolver(self)
        return ODE_solver

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, new_state):
        self._state = new_state
        if hasattr(self, "integrator"):
            self.integrator.state = self.state

    def create_CasADi_integrator(self, max_step):
        state_symbolic = rg.array_symb(self.system.dim_state, literal="x")
        action_symbolic = rg.array_symb(self.system.dim_inputs, literal="u")
        time = rg.array_symb((1, 1), literal="t")

        ODE = self.system.compute_state_dynamics(
            time, state_symbolic, action_symbolic, _native_dim=True
        )
        DAE = {"x": state_symbolic, "p": action_symbolic, "ode": ODE}

        # options = {"tf": max_step, "atol": self.atol, "rtol": self.rtol}
        options = {"tf": max_step}
        integrator = casadi.integrator("intg", "rk", DAE, 0, max_step)

        # integrator = casadi.integrator("intg", "rk", DAE, 0, max_step)

        return integrator



"""Contains Enum class representing different events that may occur inside the main loop."""

from enum import Enum, auto


class Event(Enum):
    """Enum representing different events that may occur inside the main loop.

    These events can be used to trigger specific actions or optimizations during the execution of the main loop.

    Attributes:
        reset_iteration: Indicates that an event occurs after resetting
            the iteration counter.
        reset_episode: Indicates that an event occurs after resetting
            the episode counter.
        compute_action: Indicates that an event occurs after computing
            an action.
        reset_simulation: Indicates that an event occurs after resetting
            the simulation.

    Usage:
        event = Event.reset_iteration
        if event == Event.reset_iteration:
            # Perform specific action or optimization after resetting the iteration counter
            pass
    """

    reset_iteration = auto()
    reset_episode = auto()
    compute_action = auto()
    reset_simulation = auto()



from types import MappingProxyType

from regelum import RegelumBase


class Metadata(RegelumBase):
    def __init__(self, content):
        self.content = MappingProxyType(content)

    def __enter__(self):
        self._metadata = self.content

    def __exit__(self, exc_type, exc_val, exc_tb):
        delattr(RegelumBase, f"_{RegelumBase.__name__}__metadata")



import regelum
from abc import abstractmethod, ABC
import numpy as np
from regelum.data_buffers import DataBuffer
from typing import Union


class StabilizationCriterion(regelum.RegelumBase, ABC):
    @abstractmethod
    def __call__(self, data_buffer: DataBuffer) -> bool:
        pass


class CoordinateMaxAbsCriterion(StabilizationCriterion):
    def __init__(self, max_abs: Union[float, np.ndarray], n_last_observations: int):
        self.max_abs = max_abs
        self.n_latest_observations = n_last_observations

    def __call__(self, data_buffer: DataBuffer) -> bool:
        data = data_buffer.to_pandas(["observation", "episode_id"])
        data["episode_id"] = data["episode_id"].astype(int)
        latest = (
            data.groupby(["episode_id"])["observation"]
            .apply(lambda x: x[-self.n_latest_observations :])
            .groupby(level=0)
            .mean()
        )

        for v in latest:
            if np.any(np.array(np.abs(v), dtype=float) >= self.max_abs):
                return False
        return True


class InvertedPendulumCriterion(CoordinateMaxAbsCriterion):
    def __init__(self):
        super().__init__(np.array([0.1, 0.1]), n_last_observations=10)


class TwoTankCriterion(CoordinateMaxAbsCriterion):
    def __init__(self):
        super().__init__(max_abs=0.1, n_last_observations=10)


class ThreeWheeledRobotKinematicCriterion(CoordinateMaxAbsCriterion):
    def __init__(self):
        super().__init__(max_abs=0.3, n_last_observations=10)


class ThreeWheeledRobotDynamicCriterion(CoordinateMaxAbsCriterion):
    def __init__(self):
        super().__init__(max_abs=0.3, n_last_observations=10)


class LunarLanderCriterion(CoordinateMaxAbsCriterion):
    def __init__(self):
        super().__init__(
            max_abs=np.array([0.2, 0.01, 0.001, np.inf, np.inf, np.inf]),
            n_last_observations=10,
        )


class KinPointCriterion(CoordinateMaxAbsCriterion):
    def __init__(self):
        super().__init__(
            max_abs=0.05,
            n_last_observations=10,
        )


class CartpoleCriterion(CoordinateMaxAbsCriterion):
    def __init__(self):
        super().__init__(max_abs=0.05, n_last_observations=10)


class NeverCriterion(StabilizationCriterion):
    def __call__(self, data_buffer: DataBuffer) -> bool:
        return False



"""Contains a tool box for parsing constraints that are injected outside."""

from . import RegelumBase
from .utils import rg
from abc import abstractmethod, ABC
import numpy as np
from itertools import groupby
from typing import List, Optional, Generator
from dataclasses import dataclass


def all_equal(iterable):
    g = groupby(iterable)
    return next(g, True) and not next(g, False)


def state_wise(func):
    def wrapped_constr(predicted_states=None, **kwargs):
        if len(predicted_states.shape) == 1:
            predicted_states = rg.force_row(predicted_states)
        assert predicted_states is not None, "states cannot be None"
        return rg.max(
            rg.vstack(
                [
                    func(**kwargs, state=predicted_states[i, :])
                    for i in range(predicted_states.shape[0])
                ]
            )
        )

    return wrapped_constr


def row_wise(with_respect_to: Optional[List[str]] = None):
    if with_respect_to is None:
        with_respect_to = []

    def row_wise_inner(constr_func):
        def wrapped_constr(**kwargs):
            if not kwargs:
                return constr_func()
            else:
                lens = [
                    v.shape[0] for k, v in kwargs.items() if k not in with_respect_to
                ]
                assert all_equal(
                    lens
                ), "Numbers of rows must be the same in order to apply row_wise decorator."
                len_single = lens.pop(0)
                return rg.force_column(
                    rg.vstack(
                        [
                            constr_func(
                                **{
                                    k: v[i, :] if k not in with_respect_to else v
                                    for k, v in kwargs.items()
                                }
                            )
                            for i in range(len_single)
                        ]
                    )
                )

        return wrapped_constr

    return row_wise_inner


class ConstraintParser(RegelumBase, ABC):
    """Base class for Constraint Parser."""

    @dataclass
    class ConstraintParameter:
        """Dataclass that represents constraint parameters."""

        name: str
        dims: tuple
        data: np.ndarray

    def __init__(self) -> None:
        """Instantiate ConstraintParser."""
        pass

    def parse_constraints(self, simulation_metadata=None):
        # if simulation_metadata is None:
        #     return {}
        # else:
        return self._parse_constraints(simulation_metadata=simulation_metadata)

    def __iter__(self) -> Generator[ConstraintParameter, None, None]:
        yield from self.constraint_parameters()

    @abstractmethod
    def _parse_constraints(self, simulation_metadata): ...

    @abstractmethod
    def constraint_parameters(self): ...

    @abstractmethod
    def constraint_function(self, **kwargs): ...


class ConstraintParserTrivial(ConstraintParser):
    """Trivial constraint parser that does nothing."""

    def _parse_constraints(self, simulation_metadata):
        return {}

    def constraint_parameters(self):
        return []

    def constraint_function(self, whatever=None):
        return -1


def assert_shape(array, shape, message):
    assert array.shape == tuple(shape), message


@state_wise
@row_wise(with_respect_to=["state"])
def linear_constraint(weights, bias, state):
    assert state is not None, "state cannot be None"
    return state @ rg.array(weights, prototype=state) + rg.array(bias, prototype=state)


@state_wise
@row_wise(with_respect_to=["state"])
def circle_constraint(coefs, radius, center, state):
    assert state is not None, "state cannot be None"
    return rg.array(radius, prototype=state) ** 2 - (
        rg.dot(
            rg.array(coefs, prototype=state),
            (state - rg.array(center, prototype=state)) ** 2,
        )
    )


class CylindricalHalfPlaneConstraintParser(ConstraintParser):
    """Constraint parser for ThreeWheeledRobotKinematic that consist of several circles and lines."""

    def __init__(
        self,
        centers: Optional[np.ndarray] = None,
        coefs: Optional[np.ndarray] = None,
        radii: Optional[np.ndarray] = None,
        weights: Optional[np.ndarray] = None,
        biases: Optional[np.ndarray] = None,
    ) -> None:
        """Instantiate CylindricalHalfPlaneConstraintParser.

        Args:
            centers (Optional[np.ndarray], optional): centers of
                circles, defaults to None
            coefs (Optional[np.ndarray], optional): circle coeficients,
                defaults to None
            radii (Optional[np.ndarray], optional): radii of circles,
                defaults to None
            weights (Optional[np.ndarray], optional): lines'
                coefficients, defaults to None
            biases (Optional[np.ndarray], optional): lines' biases,
                defaults to None
        """
        self.radii = np.array(radii) if radii is not None else radii
        self.centers = np.array(centers) if centers is not None else centers
        self.coefs = np.array(coefs) if coefs is not None else coefs
        self.weights = np.array(weights) if weights is not None else weights
        self.biases = np.array(biases) if biases is not None else biases

    def _parse_constraints(self, simulation_metadata=None):
        return {
            "radii": self.radii,
            "centers": self.centers,
            "coefs": self.coefs,
            "weights": self.weights,
            "biases": self.biases,
        }

    def constraint_parameters(self):
        return [
            self.ConstraintParameter(name, data.shape, data)
            for name, data in [
                ["radii", self.radii],
                ["centers", self.centers],
                ["coefs", self.coefs],
                ["weights", self.weights],
                ["biases", self.biases],
            ]
            if data is not None
        ]

    def constraint_function(
        self,
        weights=None,
        biases=None,
        radii=None,
        centers=None,
        coefs=None,
        predicted_states=None,
    ):
        weights = weights if weights is not None else self.weights
        biases = biases if biases is not None else self.biases
        radii = radii if radii is not None else self.radii
        centers = centers if centers is not None else self.centers
        coefs = coefs if coefs is not None else self.coefs

        linear_constraint_values = None
        circle_constraint_values = None
        if weights is not None and biases is not None:
            linear_constraint_values = linear_constraint(
                weights=weights, bias=biases, predicted_states=predicted_states
            )
        if radii is not None and centers is not None and coefs is not None:
            circle_constraint_values = circle_constraint(
                coefs=coefs,
                radius=radii,
                center=centers,
                predicted_states=predicted_states,
            )
        constraint_values = rg.vstack(
            [
                val
                for val in [linear_constraint_values, circle_constraint_values]
                if val is not None
            ]
        )
        return rg.max(constraint_values)



"""Contains BatchSamplers for data buffers."""

from abc import ABC, abstractmethod
from typing import Optional, List, Dict, Union
from regelum.typing import RgArrayType, RgArray
import numpy as np
import torch


class BatchSampler(ABC):
    """Base class for batch samplers."""

    def __init__(
        self,
        data_buffer,
        keys: Optional[List[str]],
        dtype: RgArrayType = torch.FloatTensor,
        device: Optional[Union[str, torch.device]] = None,
        fill_na: Optional[float] = 0.0,
    ):
        """Instantiate a BatchSampler.

        Args:
            data_buffer (DataBuffer): Data Buffer instance
            keys (Optional[List[str]]): keys to sample
            dtype (RgArrayType, optional): dtype for sample, can be
                either cs.DM, np.array, torch.Tensor, defaults to
                torch.FloatTensor
            device (Optional[Union[str, torch.device]], optional):
                device for sampling, needed for torch.FloatTensor
                defaults to None
            fill_na (Optional[float], optional, defaults to 0.0): fill
                value for np.nan, defaults to 0.0
        """
        self.keys = keys
        self.dtype = dtype
        self.data_buffer = data_buffer
        self.data_buffer.set_indexing_rules(
            keys=self.keys, dtype=self.dtype, device=device, fill_na=fill_na
        )
        self.len_data_buffer = len(self.data_buffer.data[self.keys[0]])
        self.device = device
        for k in self.keys:
            assert self.len_data_buffer == len(
                self.data_buffer.data[k]
            ), "All keys should have the same length in Data Buffer"

    def __iter__(self):
        if self.stop_iteration_criterion():
            self.nullify_sampler()
        return self

    def __next__(self):
        if self.stop_iteration_criterion():
            raise StopIteration
        return self.next()

    @abstractmethod
    def next(self) -> Dict[str, RgArray]:
        pass

    @abstractmethod
    def nullify_sampler(self) -> None:
        pass

    @abstractmethod
    def stop_iteration_criterion(self) -> bool:
        pass


class RollingBatchSampler(BatchSampler):
    """Batch sampler for rolling batches."""

    def __init__(
        self,
        mode: str,
        data_buffer,
        keys: Optional[List[str]],
        batch_size: Optional[int] = None,
        n_batches: Optional[int] = None,
        dtype: RgArrayType = torch.FloatTensor,
        device: Optional[Union[str, torch.device]] = None,
        fill_na: Optional[float] = 0.0,
    ):
        """Instantiate a RollingBatchSampler.

        Args:
            mode (str): mode for batch sampling. Can be either
                'uniform', 'backward', 'forward', 'full'. 'forward' for
                sampling of rolling batches from the beginning of
                DataBuffer. 'backward' for sampling of rolling batches
                from the end of DataBuffer. 'uniform' for sampling
                random uniformly batches. 'full' for sampling the full
                DataBuffer
            data_buffer (DataBuffer): DataBuffer instance
            keys (Optional[List[str]]): DataBuffer keys for sampling
            batch_size (Optional[int], optional): batch size, needed for
                'uniform', 'backward', 'forward', defaults to None
            n_batches (Optional[int], optional): how many batches to
                sample, can be used for all modes. Note that sampling
                procedure stops in case if DataBuffer is exhausted for
                'forward' and 'backward' modes,  defaults to None
            dtype (RgArrayType, optional): dtype for sampling, can be
                either of cs.DM, np.array, torch.Tensor, defaults to
                torch.FloatTensor
            device (Optional[Union[str, torch.device]], optional):
                device to sample from, defaults to None
            fill_na (Optional[float], optional): fill value for np.nan,
                defaults to 0.0
        """
        if batch_size is None and mode in ["uniform", "backward", "forward"]:
            raise ValueError(
                "batch_size should not be None for modes ['uniform', 'backward', 'forward']"
            )
        assert mode in [
            "uniform",
            "backward",
            "forward",
            "full",
        ], "mode should be one of ['uniform', 'backward', 'forward', 'full']"
        assert not (
            n_batches is None and (mode == "uniform" or mode == "full")
        ), "'uniform' and 'full' mode are not avaliable for n_batches == None"

        BatchSampler.__init__(
            self,
            data_buffer=data_buffer,
            keys=keys,
            dtype=dtype,
            device=device,
            fill_na=fill_na,
        )
        self.mode = mode
        self.batch_size = batch_size
        self.n_batches = n_batches
        self.n_batches_sampled: int
        self.nullify_sampler()

    def nullify_sampler(self) -> None:
        self.n_batches_sampled = 0
        if self.mode == "forward":
            self.batch_ids = np.arange(self.batch_size, dtype=int)
        elif self.mode == "backward":
            self.batch_ids = np.arange(
                self.len_data_buffer - self.batch_size,
                self.len_data_buffer,
                dtype=int,
            )
        elif self.mode == "uniform":
            self.batch_ids = np.random.randint(
                low=0,
                high=max(self.len_data_buffer - self.batch_size, 1),
            ) + np.arange(self.batch_size, dtype=int)
        elif self.mode == "full":
            self.batch_ids = np.arange(self.len_data_buffer, dtype=int)
        else:
            raise ValueError("mode should be one of ['uniform', 'backward', 'forward']")

    def stop_iteration_criterion(self) -> bool:
        if self.mode != "full":
            if self.len_data_buffer <= self.batch_size:
                return True
        if self.mode == "forward":
            return (
                self.batch_ids[-1] >= len(self.data_buffer)
                or self.n_batches == self.n_batches_sampled
            )
        elif self.mode == "backward":
            return self.batch_ids[0] <= 0 or self.n_batches == self.n_batches_sampled
        elif self.mode == "uniform" or self.mode == "full":
            return self.n_batches == self.n_batches_sampled
        else:
            raise ValueError(
                "mode should be one of ['uniform', 'backward', 'forward', 'full']"
            )

    def next(self) -> Dict[str, RgArray]:
        batch = self.data_buffer[self.batch_ids]
        if self.mode == "forward":
            self.batch_ids += 1
        elif self.mode == "backward":
            self.batch_ids -= 1
        elif self.mode == "uniform":
            self.batch_ids = np.random.randint(
                low=0, high=self.len_data_buffer - self.batch_size
            ) + np.arange(self.batch_size, dtype=int)

        # for self.mode == "full" we should not update batch_ids as they are constant for full mode
        # i. e. self.batch_ids == np.arange(self.len_data_buffer, dtype=int)

        self.n_batches_sampled += 1
        return batch


class EpisodicSampler(BatchSampler):
    """Samples the whole episodes from DataBuffer."""

    def __init__(
        self,
        data_buffer,
        keys: List[str],
        dtype: RgArrayType = torch.FloatTensor,
        device: Optional[Union[str, torch.device]] = None,
        fill_na: Optional[float] = 0.0,
    ):
        """Instantiate a EpisodicSampler.

        Args:
            data_buffer (DataBuffer): instance of DataBuffer
            keys (List[str]): keys for sampling
            dtype (RgArrayType, optional): batch dtype for sampling, can
                be either of cs.DM, np.array, torch.Tensor, defaults to
                torch.FloatTensor
            device (Optional[Union[str, torch.device]], optional):
                torch.Tensor device for sampling, defaults to None
            fill_na (Optional[float], optional): fill value for np.nan,
                defaults to 0.0
        """
        BatchSampler.__init__(
            self,
            data_buffer=data_buffer,
            keys=keys,
            dtype=dtype,
            device=device,
            fill_na=fill_na,
        )
        self.nullify_sampler()

    def nullify_sampler(self) -> None:
        self.episode_ids = (
            self.data_buffer.to_pandas(keys=["episode_id"])
            .astype(int)
            .values.reshape(-1)
        )
        self.max_episode_id = max(self.episode_ids)
        self.cur_episode_id = min(self.episode_ids) - 1
        self.idx_batch = -1

    def stop_iteration_criterion(self) -> bool:
        return self.cur_episode_id >= self.max_episode_id

    def get_episode_batch_ids(self, episode_id) -> np.array:
        return np.arange(len(self.data_buffer), dtype=int)[
            self.episode_ids == episode_id
        ]

    def next(self) -> Dict[str, RgArray]:
        self.cur_episode_id += 1
        batch_ids = self.get_episode_batch_ids(self.cur_episode_id)
        return self.data_buffer[batch_ids]


class RandomSampler(BatchSampler):
    def __init__(
        self,
        data_buffer,
        keys: List[str],
        dtype: RgArrayType = torch.FloatTensor,
        device: Optional[Union[str, torch.device]] = None,
        fill_na: Optional[float] = 0.0,
        split_size: int = 100,
        n_splits_per_episode: int = 1,
        n_batches: int = 10,
    ):
        """Instantiate a EpisodicSampler.

        Args:
            data_buffer (DataBuffer): instance of DataBuffer
            keys (List[str]): keys for sampling
            dtype (RgArrayType, optional): batch dtype for sampling, can
                be either of cs.DM, np.array, torch.Tensor, defaults to
                torch.FloatTensor
            device (Optional[Union[str, torch.device]], optional):
                torch.Tensor device for sampling, defaults to None
            fill_na (Optional[float], optional): fill value for np.nan,
                defaults to 0.0
        """
        BatchSampler.__init__(
            self,
            data_buffer=data_buffer,
            keys=keys,
            dtype=dtype,
            device=device,
            fill_na=fill_na,
        )
        self.split_size = split_size
        self.n_splits_per_episode = n_splits_per_episode
        self.n_batches = n_batches
        self.nullify_sampler()

    def nullify_sampler(self) -> None:
        self.episode_ids = (
            self.data_buffer.to_pandas(keys=["episode_id"])
            .astype(int)
            .values.reshape(-1)
        )
        self.unique_episode_ids = np.unique(self.episode_ids)
        self.idx_batch = 0

    def stop_iteration_criterion(self) -> bool:
        return self.idx_batch >= self.n_batches

    @staticmethod
    def get_split_indexes_for_episode(
        episode_size, max_n_splits_per_episode, max_split_size, episode_start_index
    ) -> list[int]:
        n_splits_per_episode = min(max_n_splits_per_episode, episode_size)
        split_gen_range_size = episode_size // n_splits_per_episode
        split_size = min(max_split_size, split_gen_range_size)

        split_ranges = np.array(
            [
                [
                    i * split_gen_range_size,
                    (
                        (i + 1) * split_gen_range_size
                        if i < n_splits_per_episode - 1
                        else episode_size
                    )
                    - split_size
                    + 1,
                ]
                for i in range(n_splits_per_episode)
            ]
        )
        ids = []
        for sr in split_ranges:
            left = np.random.randint(low=sr[0], high=sr[1])
            right = left + split_size
            ids.extend(range(episode_start_index + left, episode_start_index + right))

        return ids

    def next(self) -> Dict[str, RgArray]:
        self.idx_batch += 1

        batch_ids = []
        for episode_id in self.unique_episode_ids:
            argwhere_episode = np.argwhere(self.episode_ids == episode_id)
            episode_start_index = argwhere_episode[0, 0]
            episode_size = len(argwhere_episode)
            batch_ids.extend(
                self.get_split_indexes_for_episode(
                    episode_size=episode_size,
                    max_n_splits_per_episode=self.n_splits_per_episode,
                    max_split_size=self.split_size,
                    episode_start_index=episode_start_index,
                )
            )

        return self.data_buffer[np.array(batch_ids)]
